# Module docstring
"""
:class:`royalnet.engineer.bullet.Bullet`\\ s for the :mod:`royalnet_console` frontend.
"""

# Special imports
from __future__ import annotations
import royalnet.royaltyping as t

# External imports
import logging
import datetime
import os
import getpass
import psutil
import royalnet.engineer as engi

# Internal imports
from .utils.message import console_message

# Special global objects
log = logging.getLogger(__name__)


# Code
class ConsoleUser(engi.User):
    def __init__(self, mag: engi.Magazine):
        super().__init__(mag)

    def __hash__(self) -> int:
        return os.getuid()

    async def name(self) -> str:
        return getpass.getuser()

    async def send_message(self, *,
                           text: str = None,
                           files: t.List[t.BinaryIO] = None) -> engi.Message:
        return await console_message(mag=self.mag, text=text, files=files)


class ConsoleChannel(engi.Channel):
    def __init__(self, mag: engi.Magazine):
        super().__init__(mag)

    def __hash__(self) -> int:
        return os.getpid()

    async def name(self) -> str:
        return psutil.Process(os.getpid()).name()

    async def users(self) -> t.List[engi.User]:
        return [self.mag.User()]

    async def send_message(self, *,
                           text: str = None,
                           files: t.List[t.BinaryIO] = None) -> engi.Message:
        return await console_message(mag=self.mag, text=text, files=files)


class ConsoleMessage(engi.Message):
    _instance_count: int = 0

    def __init__(self, mag: engi.Magazine, _text: str, _timestamp: datetime.datetime = None):
        super().__init__(mag)
        self._text: str = _text
        self._timestamp: datetime.datetime = _timestamp or datetime.datetime.now()
        self._instance_number: int = self._instance_count
        self._instance_count += 1

    def __hash__(self) -> int:
        return self._instance_number

    async def text(self) -> str:
        return self._text

    async def timestamp(self) -> datetime.datetime:
        return self._timestamp

    async def channel(self) -> engi.Channel:
        return self.mag.Channel()

    async def send_reply(self, *,
                         text: str = None,
                         files: t.List[t.BinaryIO] = None) -> engi.Message:
        return await console_message(mag=self.mag, text=text, files=files)


# Objects exported by this module
__all__ = (
    "ConsoleUser",
    "ConsoleChannel",
    "ConsoleMessage",
)



# Module docstring
"""
The :class:`royalnet.engineer.magazine.Magazine` for the :mod:`royalnet_console` frontend.
"""

# Special imports
from __future__ import annotations
import royalnet.royaltyping as t

# External imports
import logging
import royalnet.engineer as engi

# Internal imports
from . import bullets

# Special global objects
log = logging.getLogger(__name__)


# Code
class ConsoleMagazine(engi.Magazine):
    _USER = bullets.ConsoleUser
    _CHANNEL = bullets.ConsoleChannel
    _MESSAGE = bullets.ConsoleMessage


# Objects exported by this module
__all__ = (
    "ConsoleMagazine",
)



# Module docstring
"""
The PDA ("main" class) for the :mod:`royalnet_console` frontend.
"""

# Special imports
from __future__ import annotations
import royalnet.royaltyping as t

# External imports
import logging
import asyncio
import royalnet.engineer as engi
import click

# Internal imports
from . import magazine

# Special global objects
log = logging.getLogger(__name__)


# Code
class ConsolePDA:
    """
    A PDA which handles :mod:`royalnet` input and output using a terminal as source.
    """

    def __init__(self):
        log.debug(f"Creating new ConsolePDA...")

        log.debug(f"Creating new magazine...")
        self.mag = magazine.ConsoleMagazine()
        """
        The :class:`royalnet.engineer.magazine.Magazine` used by this PDA.
        """

        self.dispenser: t.Optional[engi.Dispenser] = None
        """
        The :class:`royalnet.engineer.dispenser.Dispenser` of this PDA.
        """

        self.conversations: t.List[engi.Conversation] = []
        """
        A :class:`list` of conversations to run before a new event is :meth:`.put` in a 
        :class:`~royalnet.engineer.dispenser.Dispenser`.
        """

    async def run(self) -> t.NoReturn:
        """
        Run the main loop of the :class:`.ConsolePDA`.
        """

        while True:
            message = click.prompt("", type=str, prompt_suffix=">>> ", show_default=False)
            log.debug(f"Received a new message: {message!r}")

            log.debug(f"Creating ConsoleMessage from: {message!r}")
            bullet = self.mag.Message(_text=message)

            log.debug(f"Putting bullet: {bullet!r}")
            await self.put_bullet(bullet=bullet)

    def register_conversation(self, conv: engi.Conversation) -> None:
        """
        Register a new conversation in the PDA.

        :param conv: The conversation to register.
        """
        log.info(f"Registering conversation: {conv!r}")
        self.conversations.append(conv)

    def unregister_conversation(self, conv: engi.Conversation) -> None:
        """
        Unregister a conversation from the PDA.

        :param conv: The conversation to unregister.
        """
        log.info(f"Unregistering conversation: {conv!r}")
        self.conversations.remove(conv)

    def register_partial(self, part: engi.PartialCommand, names: t.List[str]) -> engi.Command:
        """
        Register a new :class:`~royalnet.engineer.command.PartialCommand` in the PDA, converting it to a
        :class:`royalnet.engineer.Command` in the process.

        :param part: The :class:`~royalnet.engineer.command.PartialCommand` to register.
        :param names: The :attr:`~royalnet.engineer.command.Command.names` to register the command with.
        :return: The resulting :class:`~royalnet.engineer.command.Command`.
        """
        log.debug(f"Completing partial: {part!r}")
        if part.syntax:
            command = part.complete(pattern=r"^{name}\s+{syntax}$", names=names)
        else:
            command = part.complete(pattern=r"^{name}$", names=names)
        self.register_conversation(command)
        return command

    async def put_bullet(self, bullet: engi.Bullet) -> None:
        """
        Insert a new bullet into the dispenser.

        :param bullet: The bullet to put in the dispenser.
        """
        if not self.dispenser:
            log.debug(f"Dispenser not found, creating one...")
            self.dispenser = engi.Dispenser()

        log.debug("Getting running loop...")
        loop = asyncio.get_running_loop()

        for conversation in self.conversations:
            log.debug(f"Creating run task for: {conversation!r}")
            loop.create_task(self.dispenser.run(conversation), name=f"{repr(conversation)}")

        log.debug("Running a event loop cycle...")
        await asyncio.sleep(0)

        log.debug(f"Putting bullet {bullet!r} in dispenser {self.dispenser!r}...")
        await self.dispenser.put(bullet)

        log.debug("Awaiting another event loop cycle...")
        await asyncio.sleep(0)


# Objects exported by this module
__all__ = (
    "ConsolePDA",
)



# Module docstring
"""
An universal function to send messages to the console.
"""

# Special imports
from __future__ import annotations
import royalnet.royaltyping as t

# External imports
import logging
import royalnet.engineer as engi
import click

# Internal imports
# from . import something

# Special global objects
log = logging.getLogger(__name__)


# Code
async def console_message(*,
                          mag: engi.Magazine,
                          text: str = None,
                          files: t.List[t.BinaryIO] = None) -> engi.Message:
    """
    Output a message to the console and return the resulting bullet.

    :param mag: The :class:`.engi.Magazine` to use when instantiating the bullet.
    :param text: The text of the message.
    :param files: A :class:`list` of files to attach to the message.
    :return: The sent :class:`.engi.Message`.
    """
    if files is None:
        files = []

    if len(files) > 0:
        raise engi.exc.NotSupportedError("Console does not allow sending files.")

    log.debug("Sending message...")
    click.echo(text)

    log.debug("Creating bullet...")
    return mag.Message(_text=text)


# Objects exported by this module
__all__ = (
    "console_message",
)



# Special imports
from __future__ import annotations
import royalnet.royaltyping as t

# External imports
import logging
import importlib
import click
import asyncio
import royalnet.engineer as engi

# Internal imports
from .pda import ConsolePDA

# Special global objects
log = logging.getLogger(__name__)


# Code
@click.command()
@click.option("-p", "--pack", "packs", multiple=True)
def main(packs: t.List[str]):
    log.debug("Creating PDA...")
    pda = ConsolePDA()

    for pack in packs:
        log.debug(f"Importing module: {pack!r}")
        try:
            pack = importlib.import_module(pack)
        except ImportError as e:
            log.error(f"Skipping {pack!r}: {e!r}")
            continue

        for attribute in dir(pack):
            log.debug(f"Getting attribute: {attribute!r}")
            value = pack.__getattribute__(attribute)
            log.debug(f"Attribute is: {value!r}")

            if isinstance(value, engi.PartialCommand):
                log.debug(f"Attribute is a PartialCommand, registering it as: {value.f.__name__!r}")
                pda.register_partial(part=value, names=[value.f.__name__])

            elif isinstance(value, engi.Conversation):
                log.debug(f"Attribute is a Conversation, registering it...")
                pda.register_conversation(conv=value)

    log.debug("Getting event loop...")
    loop = asyncio.get_event_loop()
    log.debug(f"Event loop is: {loop!r}")

    log.debug("Running the PDA until interrupted...")
    try:
        loop.run_until_complete(pda.run())
    except KeyboardInterrupt:
        log.debug("Got an interrupt, shutting down...")
        exit(0)

    log.fatal("PDA stopped unexpectedly, shutting down...")
    exit(1)


if __name__ == "__main__":
    main()
